/*--------------------------------------------------------------------

  Title       : RC5 Interrupt driven Decoder
  Filename    : rc_decoder.c 
   
  Copyright @ BobTheBass 2011
    
--------------------------------------------------------------------*/

#include <htc.h>
#include "rc5_decoder.h"

/*--- RC5 Decoder state machine ---*/

enum RC5_STATE{START_S, CALIBRATE_S, DATA_S};

/*--- RC5 code structure ---*/

volatile RC5_CODE RC5_Code;

/*--- RC5 decoder structure ---*/

typedef struct
  {
  enum RC5_STATE State; 
  unsigned char bit_time;
  unsigned char pulse_t1;
  unsigned char tolerance;
  unsigned char last_bit;
  unsigned char bit_counter;
  unsigned int rc_data;
  }DECODER;

volatile DECODER Decoder;

/*--- Initialise rc5 decoder interrupt ---*/

void init_rc5_decoder(void)
  {
  RC5_Code.valid = 0;
  Decoder.State = START_S;

  ANSEL &= 0xfb;  /* RA2 as digital I/O */
  TRISA2 = 1;     /* RA2 as input */

/* Timer should have an overflow period of greater than 5mS and less than 15mS */
/* Overflow period = (1 / (oscfreq / 4)) * 256 * prescaler */ 

  //OPTION = 0x04;  /* Pull ups on, Timer 1:32 prescaler for 4MHz oscillator */

  OPTION = 0x05;  /* Pull ups on, Timer 1:64 prescaler for 8MHz oscillator */

  INTF = 0;   /* Clear interrupt flag */
  INTE = 1;   /* Enable External interrupt pin (RA2 on 690) */
 
  T0IF = 0;   /* Clear interrupt flag */
  T0IE = 0;   /* Timer_0 overflow interrupt disable */

  GIE = 1;    /* Global Interrupt Enable */ 
  } 

/*--- Rc5 interrupt ---*/

void interrupt isr_vector(void)
  {
  if(INTF)
    {
    switch(Decoder.State)
      {
      case START_S:     TMR0 = 0;
                        T0IF = 0;
                        T0IE = 1; 
                        Decoder.bit_counter = 0;
                        Decoder.rc_data = 0;
                        Decoder.State = CALIBRATE_S;
                        break;

      case CALIBRATE_S: Decoder.pulse_t1 = TMR0;  /* Calibrate bit time */
                        TMR0 = 0; 
                        Decoder.tolerance = Decoder.pulse_t1 / 8; 
                        Decoder.last_bit = 1;  
                        Decoder.State = DATA_S;
                        break;
 
      case DATA_S:      Decoder.bit_time = TMR0;
                        TMR0 = 0;   

                        if(Decoder.bit_time <= (Decoder.pulse_t1 + Decoder.tolerance)){
                          if(Decoder.bit_time >= (Decoder.pulse_t1 - Decoder.tolerance)){
                            Decoder.rc_data <<= 1;
                            Decoder.bit_counter++;
                            if(Decoder.last_bit == 1){                   
                              Decoder.rc_data |= 1;
                              }
                            else{
                              Decoder.rc_data &= 0xffe;
                              }
                            }
                          }
                        else if(Decoder.bit_time <= ((Decoder.pulse_t1 + (Decoder.pulse_t1 >> 1)) + Decoder.tolerance)){
                          if(Decoder.bit_time >= ((Decoder.pulse_t1 + (Decoder.pulse_t1 >> 1)) - Decoder.tolerance)){
                            Decoder.rc_data <<= 1;
                            Decoder.bit_counter++;
                            Decoder.rc_data &= 0xffe;
                            if(Decoder.last_bit == 1){
                              Decoder.last_bit = 0;
                              }
                            else{
                              Decoder.rc_data <<= 1;
                              Decoder.bit_counter++;
                              Decoder.rc_data |= 1;
                              Decoder.last_bit = 1;
                              }
                            }
                          }
                        else if(Decoder.bit_time <= ((Decoder.pulse_t1 << 1) + Decoder.tolerance)){
                          if(Decoder.bit_time >= ((Decoder.pulse_t1 << 1) - Decoder.tolerance)){
                            Decoder.rc_data <<= 1;
                            Decoder.bit_counter++;
                            Decoder.rc_data &= 0xffe;
                            Decoder.rc_data <<= 1;
                            Decoder.bit_counter++;
                            Decoder.rc_data |= 1;
                            Decoder.last_bit = 1;
                            }
                          }
                        else{
                          T0IF = 1;
                          }  

                        break;
    
      default:          T0IF = 1;
                        break;
      }

    INTF = 0;
    }

  if(T0IF)
    { 
    if(Decoder.bit_counter == 0x0b){
      Decoder.rc_data <<= 1;
      Decoder.bit_counter++;
      Decoder.rc_data &= 0xffe;
      }

    if(Decoder.bit_counter == 0x0c)
      {
      RC5_Code.address = (Decoder.rc_data >> 6) & 0x001f;
      RC5_Code.command = Decoder.rc_data & 0x003f;
      RC5_Code.toggle = (Decoder.rc_data >> 11) & 0x0001; 
      RC5_Code.valid = 1;
      }
    
    T0IF = 0;
    T0IE = 0;
    Decoder.State = START_S; 
    }
  }

/*--- End of File ---*/
